"""
Storage SOP Class UIDs auto-generated by generate_storage_sopclass_uids.py
"""

from pydicom.uid import UID

MediaStorageDirectoryStorage = UID(
    '1.2.840.10008.1.3.10')
ComputedRadiographyImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.1')
DigitalXRayImageStorageForPresentation = UID(
    '1.2.840.10008.5.1.4.1.1.1.1')
DigitalXRayImageStorageForProcessing = UID(
    '1.2.840.10008.5.1.4.1.1.1.1.1')
DigitalMammographyXRayImageStorageForPresentation = UID(
    '1.2.840.10008.5.1.4.1.1.1.2')
DigitalMammographyXRayImageStorageForProcessing = UID(
    '1.2.840.10008.5.1.4.1.1.1.2.1')
DigitalIntraOralXRayImageStorageForPresentation = UID(
    '1.2.840.10008.5.1.4.1.1.1.3')
DigitalIntraOralXRayImageStorageForProcessing = UID(
    '1.2.840.10008.5.1.4.1.1.1.3.1')
EncapsulatedPDFStorage = UID(
    '1.2.840.10008.5.1.4.1.1.104.1')
EncapsulatedCDAStorage = UID(
    '1.2.840.10008.5.1.4.1.1.104.2')
EncapsulatedSTLStorage = UID(
    '1.2.840.10008.5.1.4.1.1.104.3')
EncapsulatedOBJStorage = UID(
    '1.2.840.10008.5.1.4.1.1.104.4')
EncapsulatedMTLStorage = UID(
    '1.2.840.10008.5.1.4.1.1.104.5')
GrayscaleSoftcopyPresentationStateStorage = UID(
    '1.2.840.10008.5.1.4.1.1.11.1')
SegmentedVolumeRenderingVolumetricPresentationStateStorage = UID(
    '1.2.840.10008.5.1.4.1.1.11.10')
MultipleVolumeRenderingVolumetricPresentationStateStorage = UID(
    '1.2.840.10008.5.1.4.1.1.11.11')
ColorSoftcopyPresentationStateStorage = UID(
    '1.2.840.10008.5.1.4.1.1.11.2')
PseudoColorSoftcopyPresentationStateStorage = UID(
    '1.2.840.10008.5.1.4.1.1.11.3')
BlendingSoftcopyPresentationStateStorage = UID(
    '1.2.840.10008.5.1.4.1.1.11.4')
XAXRFGrayscaleSoftcopyPresentationStateStorage = UID(
    '1.2.840.10008.5.1.4.1.1.11.5')
GrayscalePlanarMPRVolumetricPresentationStateStorage = UID(
    '1.2.840.10008.5.1.4.1.1.11.6')
CompositingPlanarMPRVolumetricPresentationStateStorage = UID(
    '1.2.840.10008.5.1.4.1.1.11.7')
AdvancedBlendingPresentationStateStorage = UID(
    '1.2.840.10008.5.1.4.1.1.11.8')
VolumeRenderingVolumetricPresentationStateStorage = UID(
    '1.2.840.10008.5.1.4.1.1.11.9')
XRayAngiographicImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.12.1')
EnhancedXAImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.12.1.1')
XRayRadiofluoroscopicImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.12.2')
EnhancedXRFImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.12.2.1')
PositronEmissionTomographyImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.128')
LegacyConvertedEnhancedPETImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.128.1')
XRay3DAngiographicImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.13.1.1')
XRay3DCraniofacialImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.13.1.2')
BreastTomosynthesisImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.13.1.3')
BreastProjectionXRayImageStorageForPresentation = UID(
    '1.2.840.10008.5.1.4.1.1.13.1.4')
BreastProjectionXRayImageStorageForProcessing = UID(
    '1.2.840.10008.5.1.4.1.1.13.1.5')
EnhancedPETImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.130')
BasicStructuredDisplayStorage = UID(
    '1.2.840.10008.5.1.4.1.1.131')
IntravascularOpticalCoherenceTomographyImageStorageForPresentation = UID(
    '1.2.840.10008.5.1.4.1.1.14.1')
IntravascularOpticalCoherenceTomographyImageStorageForProcessing = UID(
    '1.2.840.10008.5.1.4.1.1.14.2')
CTImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.2')
EnhancedCTImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.2.1')
LegacyConvertedEnhancedCTImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.2.2')
NuclearMedicineImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.20')
CTDefinedProcedureProtocolStorage = UID(
    '1.2.840.10008.5.1.4.1.1.200.1')
CTPerformedProcedureProtocolStorage = UID(
    '1.2.840.10008.5.1.4.1.1.200.2')
ProtocolApprovalStorage = UID(
    '1.2.840.10008.5.1.4.1.1.200.3')
UltrasoundMultiFrameImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.3.1')
ParametricMapStorage = UID(
    '1.2.840.10008.5.1.4.1.1.30')
MRImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.4')
EnhancedMRImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.4.1')
MRSpectroscopyStorage = UID(
    '1.2.840.10008.5.1.4.1.1.4.2')
EnhancedMRColorImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.4.3')
LegacyConvertedEnhancedMRImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.4.4')
RTImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.481.1')
RTPhysicianIntentStorage = UID(
    '1.2.840.10008.5.1.4.1.1.481.10')
RTSegmentAnnotationStorage = UID(
    '1.2.840.10008.5.1.4.1.1.481.11')
RTRadiationSetStorage = UID(
    '1.2.840.10008.5.1.4.1.1.481.12')
CArmPhotonElectronRadiationStorage = UID(
    '1.2.840.10008.5.1.4.1.1.481.13')
TomotherapeuticRadiationStorage = UID(
    '1.2.840.10008.5.1.4.1.1.481.14')
RoboticArmRadiationStorage = UID(
    '1.2.840.10008.5.1.4.1.1.481.15')
RTRadiationRecordSetStorage = UID(
    '1.2.840.10008.5.1.4.1.1.481.16')
RTRadiationSalvageRecordStorage = UID(
    '1.2.840.10008.5.1.4.1.1.481.17')
TomotherapeuticRadiationRecordStorage = UID(
    '1.2.840.10008.5.1.4.1.1.481.18')
CArmPhotonElectronRadiationRecordStorage = UID(
    '1.2.840.10008.5.1.4.1.1.481.19')
RTDoseStorage = UID(
    '1.2.840.10008.5.1.4.1.1.481.2')
RoboticRadiationRecordStorage = UID(
    '1.2.840.10008.5.1.4.1.1.481.20')
RTStructureSetStorage = UID(
    '1.2.840.10008.5.1.4.1.1.481.3')
RTBeamsTreatmentRecordStorage = UID(
    '1.2.840.10008.5.1.4.1.1.481.4')
RTPlanStorage = UID(
    '1.2.840.10008.5.1.4.1.1.481.5')
RTBrachyTreatmentRecordStorage = UID(
    '1.2.840.10008.5.1.4.1.1.481.6')
RTTreatmentSummaryRecordStorage = UID(
    '1.2.840.10008.5.1.4.1.1.481.7')
RTIonPlanStorage = UID(
    '1.2.840.10008.5.1.4.1.1.481.8')
RTIonBeamsTreatmentRecordStorage = UID(
    '1.2.840.10008.5.1.4.1.1.481.9')
DICOSCTImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.501.1')
DICOSDigitalXRayImageStorageForPresentation = UID(
    '1.2.840.10008.5.1.4.1.1.501.2.1')
DICOSDigitalXRayImageStorageForProcessing = UID(
    '1.2.840.10008.5.1.4.1.1.501.2.2')
DICOSThreatDetectionReportStorage = UID(
    '1.2.840.10008.5.1.4.1.1.501.3')
DICOS2DAITStorage = UID(
    '1.2.840.10008.5.1.4.1.1.501.4')
DICOS3DAITStorage = UID(
    '1.2.840.10008.5.1.4.1.1.501.5')
DICOSQuadrupoleResonanceStorage = UID(
    '1.2.840.10008.5.1.4.1.1.501.6')
UltrasoundImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.6.1')
EnhancedUSVolumeStorage = UID(
    '1.2.840.10008.5.1.4.1.1.6.2')
EddyCurrentImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.601.1')
EddyCurrentMultiFrameImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.601.2')
RawDataStorage = UID(
    '1.2.840.10008.5.1.4.1.1.66')
SpatialRegistrationStorage = UID(
    '1.2.840.10008.5.1.4.1.1.66.1')
SpatialFiducialsStorage = UID(
    '1.2.840.10008.5.1.4.1.1.66.2')
DeformableSpatialRegistrationStorage = UID(
    '1.2.840.10008.5.1.4.1.1.66.3')
SegmentationStorage = UID(
    '1.2.840.10008.5.1.4.1.1.66.4')
SurfaceSegmentationStorage = UID(
    '1.2.840.10008.5.1.4.1.1.66.5')
TractographyResultsStorage = UID(
    '1.2.840.10008.5.1.4.1.1.66.6')
RealWorldValueMappingStorage = UID(
    '1.2.840.10008.5.1.4.1.1.67')
SurfaceScanMeshStorage = UID(
    '1.2.840.10008.5.1.4.1.1.68.1')
SurfaceScanPointCloudStorage = UID(
    '1.2.840.10008.5.1.4.1.1.68.2')
SecondaryCaptureImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.7')
MultiFrameSingleBitSecondaryCaptureImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.7.1')
MultiFrameGrayscaleByteSecondaryCaptureImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.7.2')
MultiFrameGrayscaleWordSecondaryCaptureImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.7.3')
MultiFrameTrueColorSecondaryCaptureImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.7.4')
VLEndoscopicImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.77.1.1')
VideoEndoscopicImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.77.1.1.1')
VLMicroscopicImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.77.1.2')
VideoMicroscopicImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.77.1.2.1')
VLSlideCoordinatesMicroscopicImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.77.1.3')
VLPhotographicImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.77.1.4')
VideoPhotographicImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.77.1.4.1')
OphthalmicPhotography8BitImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.77.1.5.1')
OphthalmicPhotography16BitImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.77.1.5.2')
StereometricRelationshipStorage = UID(
    '1.2.840.10008.5.1.4.1.1.77.1.5.3')
OphthalmicTomographyImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.77.1.5.4')
WideFieldOphthalmicPhotographyStereographicProjectionImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.77.1.5.5')
WideFieldOphthalmicPhotography3DCoordinatesImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.77.1.5.6')
OphthalmicOpticalCoherenceTomographyEnFaceImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.77.1.5.7')
OphthalmicOpticalCoherenceTomographyBscanVolumeAnalysisStorage = UID(
    '1.2.840.10008.5.1.4.1.1.77.1.5.8')
VLWholeSlideMicroscopyImageStorage = UID(
    '1.2.840.10008.5.1.4.1.1.77.1.6')
LensometryMeasurementsStorage = UID(
    '1.2.840.10008.5.1.4.1.1.78.1')
AutorefractionMeasurementsStorage = UID(
    '1.2.840.10008.5.1.4.1.1.78.2')
KeratometryMeasurementsStorage = UID(
    '1.2.840.10008.5.1.4.1.1.78.3')
SubjectiveRefractionMeasurementsStorage = UID(
    '1.2.840.10008.5.1.4.1.1.78.4')
VisualAcuityMeasurementsStorage = UID(
    '1.2.840.10008.5.1.4.1.1.78.5')
SpectaclePrescriptionReportStorage = UID(
    '1.2.840.10008.5.1.4.1.1.78.6')
OphthalmicAxialMeasurementsStorage = UID(
    '1.2.840.10008.5.1.4.1.1.78.7')
IntraocularLensCalculationsStorage = UID(
    '1.2.840.10008.5.1.4.1.1.78.8')
MacularGridThicknessAndVolumeReportStorage = UID(
    '1.2.840.10008.5.1.4.1.1.79.1')
OphthalmicVisualFieldStaticPerimetryMeasurementsStorage = UID(
    '1.2.840.10008.5.1.4.1.1.80.1')
OphthalmicThicknessMapStorage = UID(
    '1.2.840.10008.5.1.4.1.1.81.1')
CornealTopographyMapStorage = UID(
    '1.2.840.10008.5.1.4.1.1.82.1')
BasicTextSRStorage = UID(
    '1.2.840.10008.5.1.4.1.1.88.11')
EnhancedSRStorage = UID(
    '1.2.840.10008.5.1.4.1.1.88.22')
ComprehensiveSRStorage = UID(
    '1.2.840.10008.5.1.4.1.1.88.33')
Comprehensive3DSRStorage = UID(
    '1.2.840.10008.5.1.4.1.1.88.34')
ExtensibleSRStorage = UID(
    '1.2.840.10008.5.1.4.1.1.88.35')
ProcedureLogStorage = UID(
    '1.2.840.10008.5.1.4.1.1.88.40')
MammographyCADSRStorage = UID(
    '1.2.840.10008.5.1.4.1.1.88.50')
KeyObjectSelectionDocumentStorage = UID(
    '1.2.840.10008.5.1.4.1.1.88.59')
ChestCADSRStorage = UID(
    '1.2.840.10008.5.1.4.1.1.88.65')
XRayRadiationDoseSRStorage = UID(
    '1.2.840.10008.5.1.4.1.1.88.67')
RadiopharmaceuticalRadiationDoseSRStorage = UID(
    '1.2.840.10008.5.1.4.1.1.88.68')
ColonCADSRStorage = UID(
    '1.2.840.10008.5.1.4.1.1.88.69')
ImplantationPlanSRStorage = UID(
    '1.2.840.10008.5.1.4.1.1.88.70')
AcquisitionContextSRStorage = UID(
    '1.2.840.10008.5.1.4.1.1.88.71')
SimplifiedAdultEchoSRStorage = UID(
    '1.2.840.10008.5.1.4.1.1.88.72')
PatientRadiationDoseSRStorage = UID(
    '1.2.840.10008.5.1.4.1.1.88.73')
PlannedImagingAgentAdministrationSRStorage = UID(
    '1.2.840.10008.5.1.4.1.1.88.74')
PerformedImagingAgentAdministrationSRStorage = UID(
    '1.2.840.10008.5.1.4.1.1.88.75')
TwelveLeadECGWaveformStorage = UID(
    '1.2.840.10008.5.1.4.1.1.9.1.1')
GeneralECGWaveformStorage = UID(
    '1.2.840.10008.5.1.4.1.1.9.1.2')
AmbulatoryECGWaveformStorage = UID(
    '1.2.840.10008.5.1.4.1.1.9.1.3')
HemodynamicWaveformStorage = UID(
    '1.2.840.10008.5.1.4.1.1.9.2.1')
CardiacElectrophysiologyWaveformStorage = UID(
    '1.2.840.10008.5.1.4.1.1.9.3.1')
BasicVoiceAudioWaveformStorage = UID(
    '1.2.840.10008.5.1.4.1.1.9.4.1')
GeneralAudioWaveformStorage = UID(
    '1.2.840.10008.5.1.4.1.1.9.4.2')
ArterialPulseWaveformStorage = UID(
    '1.2.840.10008.5.1.4.1.1.9.5.1')
RespiratoryWaveformStorage = UID(
    '1.2.840.10008.5.1.4.1.1.9.6.1')
MultichannelRespiratoryWaveformStorage = UID(
    '1.2.840.10008.5.1.4.1.1.9.6.2')
RoutineScalpElectroencephalogramWaveformStorage = UID(
    '1.2.840.10008.5.1.4.1.1.9.7.1')
ElectromyogramWaveformStorage = UID(
    '1.2.840.10008.5.1.4.1.1.9.7.2')
ElectrooculogramWaveformStorage = UID(
    '1.2.840.10008.5.1.4.1.1.9.7.3')
SleepElectroencephalogramWaveformStorage = UID(
    '1.2.840.10008.5.1.4.1.1.9.7.4')
BodyPositionWaveformStorage = UID(
    '1.2.840.10008.5.1.4.1.1.9.8.1')
ContentAssessmentResultsStorage = UID(
    '1.2.840.10008.5.1.4.1.1.90.1')
RTBrachyApplicationSetupDeliveryInstructionStorage = UID(
    '1.2.840.10008.5.1.4.34.10')
RTBeamsDeliveryInstructionStorage = UID(
    '1.2.840.10008.5.1.4.34.7')
HangingProtocolStorage = UID(
    '1.2.840.10008.5.1.4.38.1')
ColorPaletteStorage = UID(
    '1.2.840.10008.5.1.4.39.1')
GenericImplantTemplateStorage = UID(
    '1.2.840.10008.5.1.4.43.1')
ImplantAssemblyTemplateStorage = UID(
    '1.2.840.10008.5.1.4.44.1')
ImplantTemplateGroupStorage = UID(
    '1.2.840.10008.5.1.4.45.1')
